#      A python library for getting Load Shedding schedules.
#      Copyright (C) 2021  Werner Pieterson
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
from datetime import datetime, time, timedelta
from enum import Enum
from typing import List, Tuple


_LOGGER = logging.getLogger(__name__)

MAX_MONTH_DAY = 31

DAY_AREA_EXTRA_INCREMENTS = [5, 9]
DAY_AREA_EXTR_INCREMENTS_STAGE_LOWER = [13]

STAGE_STARTING_AREAS = {
    1: 1,
    2: 9,
    3: 13,
    4: 5,
    5: 2,
    6: 10,
    7: 14,
    8: 6
}

TIME_SLOT_HOURS = 2
TIME_SLOT_MINUTES = 30


class CoCTError(Exception):
    pass


class Area(Enum):
    AREA_1 = 1
    AREA_2 = 2
    AREA_3 = 3
    AREA_4 = 4
    AREA_5 = 5
    AREA_6 = 6
    AREA_7 = 7
    AREA_8 = 8
    AREA_9 = 9
    AREA_10 = 10
    AREA_11 = 11
    AREA_12 = 12
    AREA_13 = 13
    AREA_14 = 14
    AREA_15 = 15
    AREA_16 = 16

    def __str__(self):
        return {
            self.AREA_1: "Area 1",
            self.AREA_2: "Area 2",
            self.AREA_3: "Area 3",
            self.AREA_4: "Area 4",
            self.AREA_5: "Area 5",
            self.AREA_6: "Area 6",
            self.AREA_7: "Area 7",
            self.AREA_8: "Area 8",
            self.AREA_9: "Area 9",
            self.AREA_10: "Area 10",
            self.AREA_11: "Area 11",
            self.AREA_12: "Area 12",
            self.AREA_13: "Area 13",
            self.AREA_14: "Area 14",
            self.AREA_15: "Area 15",
            self.AREA_16: "Area 16",
        }.get(self, "Unknown Area")


class Stage(Enum):
    NO_LOAD_SHEDDING = 0
    STAGE_1 = 1
    STAGE_2 = 2
    STAGE_3 = 3
    STAGE_4 = 4
    STAGE_5 = 5
    STAGE_6 = 6
    STAGE_7 = 7
    STAGE_8 = 8


SLOT = datetime.now().replace(year=1, month=1, day=1, second=0, microsecond=0)
TIMESLOTS: List[List] = [
    (SLOT.replace(hour=0, minute=0), SLOT.replace(hour=2, minute=30)),
    (SLOT.replace(hour=2, minute=0), SLOT.replace(hour=4, minute=30)),
    (SLOT.replace(hour=4, minute=0), SLOT.replace(hour=6, minute=30)),
    (SLOT.replace(hour=6, minute=0), SLOT.replace(hour=8, minute=30)),
    (SLOT.replace(hour=8, minute=0), SLOT.replace(hour=10, minute=30)),
    (SLOT.replace(hour=10, minute=0), SLOT.replace(hour=12, minute=30)),
    (SLOT.replace(hour=12, minute=0), SLOT.replace(hour=14, minute=30)),
    (SLOT.replace(hour=14, minute=0), SLOT.replace(hour=16, minute=30)),
    (SLOT.replace(hour=16, minute=0), SLOT.replace(hour=18, minute=30)),
    (SLOT.replace(hour=18, minute=0), SLOT.replace(hour=20, minute=30)),
    (SLOT.replace(hour=20, minute=0), SLOT.replace(hour=22, minute=30)),
    (SLOT.replace(hour=22, minute=0), SLOT.replace(hour=0, minute=30)),
]


def get_area_codes_by_time_slot(stage, day, time_slot):
    day = _clip_day_to_group(day)
    area_code_acc = _get_area_code_accumulation_day_start(stage, day) + time_slot
    if stage > 4:
        area_code_acc += 1
    area_code = _nomalize_area_code(stage, area_code_acc)
    area_codes = [area_code]

    if stage == 4 and time_slot == 4 and day == 15:
        area_codes = []

    if stage > 1:
        area_codes = area_codes + get_area_codes_by_time_slot(stage - 1, day, time_slot)

    return area_codes


def get_time_slots_by_area_code(stage, day, area):
    time_slots = []
    for i in range(len(TIMESLOTS)):
        areas = get_area_codes_by_time_slot(stage, day, i + 1)
        if area in areas:
            time_slots.append(TIMESLOTS[i])
            continue

    return time_slots


def get_next_time_slot(stage: int, area_code: int) -> Tuple:
    # result = {"slot": None, "day": None, "date": None}

    if stage < 1 or stage > len(Stage):
        logging.debug("get_next_time_slot() stage out of bounds")
        return ()

    if area_code < 1 or area_code > len(Area):
        logging.debug("get_next_time_slot() area_code out of bounds")
        return ()

    d = datetime.now()  # + timedelta(hours=2)
    from_hour = d.hour
    from_day = d.day

    slot = None
    day = from_day
    day_accum = 0

    while not slot:
        slot = get_next_time_slot_in_day(
            stage, day, area_code, from_hour if day == from_day else -1
        )

        if not slot:
            if day >= MAX_MONTH_DAY:
                day = 1
            else:
                day += 1

            day_accum += 1

    return TIMESLOTS[slot - 1]

    new_date = datetime(d.year, d.month, d.day, get_time_slot_hour(slot), 0, 0)
    new_date = new_date + timedelta(days=day_accum)

    result["slot"] = slot
    result["day"] = day
    result["date"] = new_date

    return result


def get_time_slot_hour(slot):
    return (slot - 1) * TIME_SLOT_HOURS


def _get_timeslot_from_hour(hour):
    is_odd_hour = (hour % TIME_SLOT_HOURS) != 0

    time_slot = hour
    if is_odd_hour:
        time_slot -= 1

    return time_slot // TIME_SLOT_HOURS + 1


def _clip_day_to_group(day):
    if day > len(Area):
        day -= len(Area)
    return day


def _get_area_code_accumulation_day_start(stage, day):
    if day <= 1:
        return 0

    day_before = day - 1
    area_code_acc = day_before * len(TIMESLOTS)

    for i in DAY_AREA_EXTRA_INCREMENTS:
        if day >= i:
            area_code_acc += 1

    if stage <= 4:
        for i in DAY_AREA_EXTR_INCREMENTS_STAGE_LOWER:
            if day >= i:
                area_code_acc += 1

    return area_code_acc


def _nomalize_area_code(stage, area_code_acc):
    area_code = area_code_acc % len(Area)
    area_code += STAGE_STARTING_AREAS[stage] - 1
    if area_code > len(Area):
        area_code -= len(Area)

    return area_code


def get_next_time_slot_in_day(stage, day, area_code, from_hour=-1):
    slots = get_time_slots_by_area_code(stage, day, area_code)

    for slot in slots:
        slot_hour = get_time_slot_hour(slot)

        if from_hour == -1 or slot_hour > from_hour:
            return slot

    return None




from enum import Enum

from .load_shedding import (
    Area, Stage, StageError, ProviderError, get_areas, get_schedule, get_stage
)
from .providers import Province, coct, eskom


class Provider(Enum):
    ESKOM = 1
    COCT = 2

    def __call__(self, *args, **kwargs):
        return {
            Provider.ESKOM: eskom.Eskom(),
            Provider.COCT: coct.CoCT(),
        }.get(self, None)

    def load(self):
        return {
            Provider.ESKOM: eskom.Eskom(),
            Provider.COCT: coct.CoCT(),
        }.get(self, None)

    def __str__(self):
        return {
            self.ESKOM: eskom.Eskom.name,
            self.COCT: coct.CoCT.name,
        }.get(self, "Unknown Provider")



#      A python library for getting Load Shedding schedules.
#      Copyright (C) 2021  Werner Pieterson
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
from datetime import datetime, timedelta, timezone
from enum import Enum
import logging
from typing import Any, Dict, List, Tuple

import certifi
import urllib3
from bs4 import BeautifulSoup


class EskomError(Exception):
    pass


class Province(Enum):
    EASTERN_CAPE = 1
    FREE_STATE = 2
    GAUTENG = 3
    KWAZULU_NATAL = 4
    LIMPOPO = 6
    MPUMALANGA = 5
    NORTH_WEST = 7
    NORTERN_CAPE = 8
    WESTERN_CAPE = 9

    def __str__(self):
        return {
            self.EASTERN_CAPE: "Eastern Cape",
            self.FREE_STATE: "Free State",
            self.GAUTENG: "Gauteng",
            self.KWAZULU_NATAL: "Kwa-Zulu Natal",
            self.LIMPOPO: "Limpopo",
            self.MPUMALANGA: "Mpumalanga",
            self.NORTH_WEST: "North West",
            self.NORTERN_CAPE: "Nortern Cape",
            self.WESTERN_CAPE: "Western Cape",
        }.get(self, "Unknown Province")


class LoadShedding:
    base_url = "https://loadshedding.eskom.co.za/LoadShedding"

    def find_suburbs(self, search_text: str, max_results: int = 10) -> List[Dict]:
        if not max_results:
            max_results = 10
        url = f"{self.base_url}/FindSuburbs?searchText={search_text}&maxResults={max_results}"
        data = _call(url)
        return json.loads(data)

    def get_municipalities(self, province: Province) -> List[Dict]:
        url = f"{self.base_url}/GetMunicipalities/?Id={province.value}"
        data = _call(url)
        return json.loads(data)

    def get_schedule(self, province: Province, suburb_id: int, stage: int) -> List[Tuple]:
        tot: int = 3252
        url = f"{self.base_url}/GetScheduleM/{suburb_id}/{stage}/{province}/{tot}"
        data = _call(url)
        return parse_schedule_data(data)

    def get_schedule_area_info(self, suburb_id: int) -> Dict:
        url = f"{self.base_url}/GetScheduleAreaInfo/?Id={suburb_id}"
        data = _call(url)
        return parse_area_info_data(data)

    def get_status(self) -> int:
        url = f"{self.base_url}/GetStatus"
        data = _call(url)
        return int(data)


def _call(url: str) -> Any:
    try:
        with urllib3.PoolManager(retries=urllib3.Retry(total=3), ca_certs=certifi.where()) as conn:
            r = conn.request('GET', url)
            if r.status != 200:
                raise urllib3.response.HTTPError(r.status)
            return r.data
    except Exception as e:
        msg = f"Eskom is unreachable. Check your connection."
        logging.debug(msg, exc_info=True)
        raise EskomError(msg) from e


def parse_area_info_data(data: str) -> Dict:
    try:
        soup = BeautifulSoup(data, "html.parser")
        items = soup.find_all("div", attrs={"class": "areaInfoItem"})
        area_info = {
            "Province": {
                "Id": int(items[0].find("input", attrs={"id": "provinceId"}).get('value').strip()),
                "Name": items[0].find("input", attrs={"id": "province"}).get('value').strip(),
            },
            "Municipality": {
                "Id": int(items[1].find("input", attrs={"id": "municipalityId"}).get('value').strip()),
                "Name": items[1].find("input", attrs={"id": "municipality"}).get('value').strip(),
            },
            "Suburb": {
                "Id": int(items[2].find("input", attrs={"id": "suburbId"}).get('value').strip()),
                "Name": items[2].find("input", attrs={"id": "suburbName"}).get('value').strip(),
            },
            "Period": items[3].contents[2].strip().split("\xa0to\xa0"),
        }
    except Exception as e:
        raise EskomError(f"Unable to parse area info data.") from e
    else:
        return area_info


def parse_schedule_data(data: str) -> List[Tuple]:
    try:
        schedule = []
        soup = BeautifulSoup(data, "html.parser")
        days_soup = soup.find_all("div", attrs={"class": "scheduleDay"})

        sast = timezone(timedelta(hours=+2), 'SAST')
        now = datetime.now(sast)
        for day in days_soup:
            date_soup = day.find("div", attrs={"class": "dayMonth"})
            date_str = date_soup.get_text().strip()
            date = datetime.strptime(date_str, "%a, %d %b")
            date = date.replace(year=now.year)

            time_soup = day.find_all("a")
            for time_tag in time_soup:
                start_str, end_str = time_tag.get_text().strip().split(" - ")
                start = datetime.strptime(start_str, "%H:%M").replace(
                    year=date.year, month=date.month, day=date.day, second=0, microsecond=0, tzinfo=sast
                )
                end = datetime.strptime(end_str, "%H:%M").replace(
                    year=date.year, month=date.month, day=date.day, second=0, microsecond=0, tzinfo=sast
                )
                if end < start:
                    end = end + timedelta(days=1)
                schedule.append((start, end))
    except Exception as e:
        raise EskomError(f"Unable to parse schedule data.") from e
    else:
        return schedule



#      A python library for getting Load Shedding schedules.
#      Copyright (C) 2021  Werner Pieterson
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
import errno
import json
import logging
import os
from datetime import datetime, timezone
from typing import Any, Dict, Final, List, Tuple

from load_shedding.providers import (
    Area, Provider, ProviderError, Stage, StageError, datetime_to_isoformat, isoformat_to_datetime
)

logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s [%(levelname)s] %(message)s",
    handlers=[
        logging.StreamHandler()
    ]
)
logger = logging.getLogger(__name__)

CACHE_TTL: Final = 86400  # 24 hours


class LoadSheddingError(Exception):
    pass


def get_areas(provider: Provider, search_text: str = None, max_results: int = None) -> List[Any]:
    try:
        areas = provider.get_areas(search_text, max_results)
    except ProviderError as e:
        msg = f"Unable to get areas from {provider.name}."
        logging.debug(e, exc_info=True)
        raise ProviderError(msg) from e
    else:
        return areas


def get_stage(provider: Provider) -> Stage:
    try:
        stage = provider.get_stage()
    except Exception as e:
        msg = f"Unable to get stage from {provider.name}."
        logging.debug(msg, exc_info=True)
        raise ProviderError(msg)
    return stage


def get_schedule(provider: Provider, area: Area, stage: Stage = None, cached: bool = True) -> List[Tuple]:
    try:
        os.makedirs('.cache')
    except OSError as e:
        if e.errno != errno.EEXIST:
            msg = f"Unable to make cache dir."
            logging.debug(msg, exc_info=True)
            raise LoadSheddingError(msg)

    if stage in [None]:
        stage = get_stage(provider)

    if stage in [Stage.UNKNOWN, Stage.NO_LOAD_SHEDDING]:
        raise StageError(f"{stage}")

    cached_schedules = {}
    cached_schedule = []
    now = datetime.now(timezone.utc)
    cache_file = f".cache/{area.province.value}_{area.id}.json"
    try:
        with open(cache_file, "r") as cache:
            for k, v in json.loads(cache.read()).items():
                cached_schedules[Stage(int(k)).value] = v

            cached_schedule = cached_schedules.get(stage.value, {})

        if not isinstance(cached_schedule, dict) or len(cached_schedule) == 0:
            """ Upgrade from previous list """
            cached = False

        if cached and cached_schedule:
            updated_at = datetime.fromisoformat(cached_schedule.get("updated_at", None))

            cache_age = (now - updated_at)
            if cache_age.total_seconds() < CACHE_TTL:
                schedule = cached_schedule.get("schedule", {})
                schedule = isoformat_to_datetime(schedule)
                return schedule
    except FileNotFoundError as e:
        logging.debug("Unable to get schedule from cache. {e}".format(e=e))

    try:
        schedule = provider.get_schedule(area=area, stage=stage)
    except Exception as e:
        if cached_schedule:
            return cached_schedule  # best effort
        msg = f"Unable to get schedule from {provider.name}."
        logging.debug(msg, exc_info=True)
        raise ProviderError(msg)
    else:
        with open(cache_file, "w") as cache:
            cached_schedules[stage.value] = {
                "updated_at": str(now),
                "schedule": datetime_to_isoformat(schedule),
            }
            cache.write(json.dumps(cached_schedules))
        return schedule


def list_to_dict(schedule: list) -> Dict:
    schedule_dict = {}
    now = datetime.now(timezone.utc)
    for item in schedule:
        start = item[0]
        end = item[1]

    schedule_dict[start.strftime("%Y-%m-%d")] = (
        start.replace(second=now.second).strftime("%H:%M"),
        end.replace(second=now.second).strftime("%H:%M"),
    )
    return schedule_dict


def dict_list_to_obj_list(data: List[Dict], T: Any) -> Any:
    objs: List = []
    for d in data:
        s = T(**d)
        objs.append(s)
    return objs



#      A python library for getting Load Shedding schedules.
#      Copyright (C) 2021  Werner Pieterson
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import abc
from datetime import datetime, timezone
from enum import Enum
from typing import List, Dict, Any, Tuple


class Area:
    def __init__(self, /, **kwargs) -> object:
        self.id: int = int(kwargs.get("id", 0))
        self.name: str = kwargs.get("name", "")
        self.municipality: str = kwargs.get("municipality", "")
        self.province: Province = Province(kwargs.get("province", 0))

    def __str__(self):
        return f"{self.name}"

    def __repr__(self):
        return f"Area({self.id}, {self.name}, {self.municipality}, {self.province})"


class Province(Enum):
    UNKNOWN = 0
    EASTERN_CAPE = 1
    FREE_STATE = 2
    GAUTENG = 3
    KWAZULU_NATAL = 4
    LIMPOPO = 6
    MPUMALANGA = 5
    NORTH_WEST = 7
    NORTERN_CAPE = 8
    WESTERN_CAPE = 9

    @staticmethod
    def from_name(name: str):
        return {
            "Eastern Cape": Province.EASTERN_CAPE,
            "Free State": Province.FREE_STATE,
            "Gauteng": Province.GAUTENG,
            "Kwa-Zulu Natal": Province.KWAZULU_NATAL,
            "Limpopo": Province.LIMPOPO,
            "Mpumalanga": Province.MPUMALANGA,
            "North West": Province.NORTH_WEST,
            "Nortern Cape": Province.NORTERN_CAPE,
            "Western Cape": Province.WESTERN_CAPE,
        }.get(name, Province.UNKNOWN)

    def __str__(self):
        return {
            self.EASTERN_CAPE: "Eastern Cape",
            self.FREE_STATE: "Free State",
            self.GAUTENG: "Gauteng",
            self.KWAZULU_NATAL: "Kwa-Zulu Natal",
            self.LIMPOPO: "Limpopo",
            self.MPUMALANGA: "Mpumalanga",
            self.NORTH_WEST: "North West",
            self.NORTERN_CAPE: "Nortern Cape",
            self.WESTERN_CAPE: "Western Cape",
        }.get(self, "Unknown Province")


class StageError(Exception):
    pass


class Stage(Enum):
    UNKNOWN = -1
    NO_LOAD_SHEDDING = 0
    STAGE_1 = 1
    STAGE_2 = 2
    STAGE_3 = 3
    STAGE_4 = 4
    STAGE_5 = 5
    STAGE_6 = 6
    STAGE_7 = 7
    STAGE_8 = 8

    def __str__(self):
        return {
            self.NO_LOAD_SHEDDING: "No Load Shedding",
            self.STAGE_1: "Stage 1",
            self.STAGE_2: "Stage 2",
            self.STAGE_3: "Stage 3",
            self.STAGE_4: "Stage 4",
            self.STAGE_5: "Stage 5",
            self.STAGE_6: "Stage 6",
            self.STAGE_7: "Stage 7",
            self.STAGE_8: "Stage 8",
        }.get(self, "Unknown Stage")

    def __repr__(self):
        return self.__str__()


def dict_list_to_obj_list(data: List[Dict], T: Any) -> List[Any]:
    objs: List[Any] = []
    for d in data:
        s = T(**d)
        objs.append(s)
    return objs


def obj_list_to_dict_list(data: List[Any]) -> List[Dict]:
    dicts: List[Dict] = []
    for d in data:
        dicts.append(d.__dict__)
    return dicts


def to_utc(data: List[Tuple]) -> List[Tuple]:
    out: List[Tuple] = []
    for d in data:
        start = d[0].astimezone(timezone.utc)
        end = d[1].astimezone(timezone.utc)
        out.append((start, end))
    return out


def datetime_to_isoformat(data: List[Tuple]) -> List[Tuple]:
    out: List[Tuple] = []
    for d in data:
        start = d[0].astimezone(timezone.utc).isoformat()
        end = d[1].astimezone(timezone.utc).isoformat()
        out.append((start, end))
    return out


def isoformat_to_datetime(data: List[Tuple]) -> List[Tuple]:
    out: List[Tuple] = []
    for d in data:
        start = datetime.fromisoformat(d[0]).astimezone(timezone.utc)
        end = datetime.fromisoformat(d[1]).astimezone(timezone.utc)
        out.append((start, end))
    return out


class ProviderError(Exception):
    pass


class Provider(abc.ABC):
    name: str = "Provider"

    @abc.abstractmethod
    def get_areas(self, search_text: str = None, max_results: int = None) -> List[Area]:
        raise NotImplemented

    @abc.abstractmethod
    def get_schedule(self, area: Area, stage: Stage) -> List[Tuple]:
        raise NotImplemented

    def get_stage(self) -> Stage:
        pass



#      A python library for getting Load Shedding schedules.
#      Copyright (C) 2021  Werner Pieterson
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import List, Tuple

from load_shedding.libs import eskom
from load_shedding.providers import dict_list_to_obj_list, ProviderError, to_utc
from load_shedding.providers import Area, Province, Stage, Provider


class Area(Area):
    def __init__(self, /, **kwargs) -> object:
        self.id: int = int(kwargs.get("Id", kwargs.get("id", 0)))
        self.name: str = kwargs.get("Name", kwargs.get("name", ""))
        self.municipality: str = kwargs.get("Municipality", kwargs.get("MunicipalityName", kwargs.get("municipality", "")))
        self.province: Province = Province.from_name(kwargs.get("ProvinceName", ""))
        self.total = kwargs.get("Total")


class Eskom(Provider):
    name = "Eskom Direct"

    def __init__(self) -> object:
        self.api = eskom.LoadShedding()

    def get_areas(self, search_text: str = None, max_results: int = None) -> List[Area]:
        try:
            areas = self.api.find_suburbs(search_text, max_results)
        except Exception as e:
            raise ProviderError(e)
        else:
            areas = dict_list_to_obj_list(areas, Area)
            return filter_empty_areas(areas)

    def get_schedule(self, area: Area, stage: Stage) -> List[Tuple]:
        try:
            schedule = self.api.get_schedule(province=area.province.value, suburb_id=area.id, stage=stage.value)
        except Exception as e:
            raise ProviderError(e)
        else:
            schedule = to_utc(schedule)
            return schedule

    def get_stage(self) -> Stage:
        try:
            status = self.api.get_status()
        except Exception as e:
            raise ProviderError(e)
        else:
            return stage_from_status(status)


def stage_from_status(status: int) -> Stage:
    return {
        1: Stage.NO_LOAD_SHEDDING,
        2: Stage.STAGE_1,
        3: Stage.STAGE_2,
        4: Stage.STAGE_3,
        5: Stage.STAGE_4,
        6: Stage.STAGE_5,
        7: Stage.STAGE_6,
        8: Stage.STAGE_7,
        9: Stage.STAGE_8,
    }.get(status, Stage.UNKNOWN)


def filter_empty_areas(suburbs: List[Area]) -> List[Area]:
    filtered: List[Area] = []
    for area in suburbs:
        if hasattr(area, "total") and not area.total:
            continue
        filtered.append(area)
    return filtered
